<?php

App::uses('BuzzBookingsAppModel', 'BuzzBookings.Model');
App::uses('BookingState', 'BuzzBookings.Model');

class Booking extends BuzzBookingsAppModel {

/**
 * @var array Belongs to associations
 */
	public $belongsTo = array(
		'Activity' => array(
			'className' => 'BuzzBookings.Activity'
		),
		'BookingState' => array(
			'className' => 'BuzzBookings.BookingState'
		)
	);

/**
 * @var array Has many associations
 */
	public $hasMany = array(
		'BookingExtra' => array(
			'className' => 'BuzzBookings.BookingExtra'
		),
		'BookingItem' => array(
			'className' => 'BuzzBookings.BookingItem'
		)
	);

/**
 * @var int Basket ID
 */
	protected $_basketId = null;

/**
 * Read for view
 *
 * @param int $id Booking ID
 * @param array $params
 * @return array
 */
	public function readForView($id, $params = []) {
		$defaults = array(
			'contain' => array(
				'Activity',
				'BookingItem' => array(
					'ActivityPackage',
					'BookingItemPlace',
					'BookingItemVoucher'
				)
			)
		);
		$params = Hash::merge($defaults, $params);
		return parent::readForView($id, $params);
	}

/**
 * Get the current basket ID from the session
 *
 * @return int
 */
	public function getBasketId() {
		if (empty($this->_basketId)) {
			$this->_basketId = CakeSession::read('Booking.id');
		}
		return $this->_basketId;
	}

/**
 * Clear the basket ID from the session
 *
 * @return bool
 */
	public function clearBasketId() {
		$this->_basketId = null;
		CakeSession::delete('Booking');
		return true;
	}

/**
 * Clears the basket of existing items
 *
 * @param bool $validate Pass as true to check it is safe to clear the basket
 * @return bool
 */
	public function clearBasket($validate = true) {
		$bookingId = $this->getBasketId();

		$this->id = $bookingId;
		if ((int)$this->field('booking_state_id') !== BookingState::UNPAID) {
			// This booking has been completed so cannot be cleared!
			return false;
		}

		$params = array(
			'BookingItem.booking_id' => $bookingId
		);
		return $this->BookingItem->deleteAll($params, true);
	}

/**
 * Get the current basket
 *
 * @return array
 */
	public function getBasket() {
		$bookingId = $this->getBasketId();
		if (empty($bookingId)) {
			return [];
		}
		$params = array(
			'conditions' => array(
				'Booking.booking_state_id' => BookingState::UNPAID
			)
		);
		return $this->readForView($bookingId, $params);
	}

/**
 * Update the current basket total.
 *
 * @return bool
 */
	protected function _updateBasketTotal() {
		$basketId = $this->getBasketId();

		$basket = $this->getBasket();

		// Calculate the total cost of the booking items.
		$total = array_sum(Hash::extract($basket, 'BookingItem.{n}.total'));

		// Add the cost of booking extras to the total.
		$total += array_sum(Hash::extract($basket, 'BookingExtra.{n}.total'));

		// Save the update total.
		$data = array(
			'id' => $basketId,
			'total_cost' => $total
		);

		return $this->save($data) !== false;
	}

/**
 * Creates a new basket (or recreates an existing basket).
 *
 * @param int $activityId ID of activity being booked
 * @param array $bookingItems Packages being booked
 * @return bool
 */
	public function createBasket($activityId, $bookingItems = []) {
		$bookingId = $this->getBasketId();

		if (!empty($bookingId)) {
			// We're editing an existing basket so need to clear it first. If
			// the basket has been completed we need to create a new one.
			if ($this->clearBasket() === false) {
				$bookingId = null;
			}
		}

		// Make sure we don't save items with zero quantity to the database.
		foreach ($bookingItems as $key => $val) {
			if ((int)$val['quantity'] === 0) {
				unset($bookingItems[$key]);
			}
		}

		// Get all the packages being booked from the database so that we can
		// create the booking item places.
		$packages = $this->Activity->ActivityPackage->find(
			'all',
			array(
				'conditions' => array(
					'ActivityPackage.id' => Hash::extract($bookingItems, '{n}.activity_package_id')
				)
			)
		);
		$packages = Hash::combine($packages, '{n}.ActivityPackage.id', '{n}.ActivityPackage');
		foreach ($bookingItems as &$bookingItem) {
			$bookingItem['BookingItemPlace'] = $this->_generateBookingItemPlace($packages, $bookingItem);
		}

		$data = array(
			'Booking' => array(
				'id' => $bookingId,
				'activity_id' => $activityId,
				'total_cost' => 0,
				'booking_state_id' => BookingState::UNPAID
			),
			'BookingItem' => $bookingItems
		);

		$result = $this->saveAssociated($data, ['deep' => true]) !== false;

		// Update the session.
		if ($result === true) {
			CakeSession::write('Booking.id', $this->id);
		} else {
			$this->clearBasketId();
		}

		return $result;
	}

/**
 * Generates a booking item place to attach to a booking item
 *
 * @param array $packages
 * @param array $bookingItem
 * @return array
 */
	protected function _generateBookingItemPlace(array $packages, array $bookingItem) {
		$package = $packages[$bookingItem['activity_package_id']];
		$data = [];

		if ((int)$package['activity_id'] === $this->Activity->getMultiExperience()) {
			$apiRefs = explode(',', $package['api_reference']);
			$apiRefs = array_map('trim', $apiRefs);

			foreach ($apiRefs as $apiRef) {
				preg_match('|^(\d+)\[([\d\.]+)\]$|', $apiRef, $matches);
				$data[] = array(
					'api_reference' => $matches[1],
					'quantity' => $bookingItem['quantity'],
					'name' => $package['name'],
					'item_unit_cost' => $matches[2]
				);
			}

		} else {
			$data[] = array(
				'api_reference' => $package['api_reference'],
				'quantity' => $bookingItem['quantity'],
				'name' => $package['name'],
				'item_unit_cost' => null
			);

		}

		return $data;
	}

/**
 * Adds a voucher to the booking.
 *
 * @param int $activityId Activity ID
 * @param array $voucher BookingItemVoucher data to save
 * @return bool
 */
	public function addVoucher($activityId, array $voucher) {
		$bookingId = $this->getBasketId();

		// Validate the voucher.
		$voucher = $this->_validateVoucher($activityId, $voucher);
		if ($voucher === false) {
			return false;
		}

		// @todo Build booking item places for the voucher.
		$bookingItemPlaces = [];
		foreach ($voucher['Voucher']['Activity'] as $activity) {
			$bookingItemPlaces[] = array(
				'api_reference' => $activity['api_package_reference'],
				'quantity' => 1,
				'name' => $voucher['Voucher']['description'],
				'item_unit_cost' => 0
			);
		}

		// Save the voucher to the booking.
		$data = array(
			'Booking' => array(
				'id' => $bookingId,
				'activity_id' => $activityId,
				'total_cost' => 0,
				'booking_state_id' => BookingState::UNPAID
			),
			'BookingItem' => array(
				array(
					'id' => null,
					'activity_package_id' => null,
					'item_unit_cost' => 0,
					'quantity' => 1,
					'total' => 0,
					'BookingItemPlace' => $bookingItemPlaces,
					'BookingItemVoucher' => $voucher
				)
			)
		);
		$result = $this->saveAssociated($data, ['deep' => true]) !== false;

		return $result;
	}

/**
 * Check that a voucher is valid
 *
 * @param int $activityId
 * @param array $data
 * @return array
 */
	protected function _validateVoucher($activityId, array $data) {
		$result = false;

		// First check that the data validates against app validation rules.
		$this->BookingItem->BookingItemVoucher->set($data);
		if ($this->BookingItem->BookingItemVoucher->validates() === false) {
			return false;
		}

		if ((int)$data['voucher_provider_id'] === $this->getApiVoucherProvider()) {
			$result = $this->_checkVoucherCodeApiCall(
				$data['voucher_code'],
				BookingApi::dateToSqlFormat($data['expiry_date'])
			);
		} else {
			$result = $this->_checkThirdPartyVoucherCode(
				$data['voucher_provider_id'],
				$data['api_voucher_id'],
				$data['voucher_code'],
				BookingApi::dateToSqlFormat($data['expiry_date'])
			);
		}

		if ($result !== false) {
			// If the voucher is valid then we need to do one final check that
			// it hasn't already been used before accepting it.
			$isAvailable = $this->_checkVoucherAvailable(
				$activityId,
				(int)$data['voucher_provider_id'],
				$data['voucher_code']
			);
			if ($isAvailable === false) {
				return false;
			}
		}

		return $result;
	}

/**
 * Check a third party voucher code.
 *
 * Checks against the stored session data (retrieved from the API) and against
 * the API.
 *
 * @param int $voucherProviderId
 * @param int $apiVoucherId
 * @param string $voucherCode
 * @param string $expiryDate Expiry date in SQL format
 * @return array|bool Returns voucher data or false if not valid
 */
	protected function _checkThirdPartyVoucherCode($voucherProviderId, $apiVoucherId, $voucherCode, $expiryDate) {
		// Check if the voucher being checked exists in our session stored
		// vouchers data.
		$voucher = $this->_getVoucherData($voucherProviderId, $apiVoucherId);
		if (empty($voucher)) {
			return false;
		}

		$result = $this->_checkThirdPartyVoucherCodeApiCall(
			$voucherCode,
			$expiryDate
		);

		if ($result !== false) {
			// Update the voucher data to be returned to include values from
			// the session data to make it easier to save to the basket.
			$result['Voucher']['participants'] = $voucher['participants'];
			$result['Voucher']['Activity'] = array(
				array(
					'api_activity_reference' => $voucher['activity_api_id'],
					'quantity' => 1
				)
			);
		}

		return $result;
	}

/**
 * API call to check vouchers (to enable unit test mocking).
 *
 * @param string $voucherCode
 * @param string $expiryDate Expiry date in SQL format
 * @return array
 */
	protected function _checkVoucherCodeApiCall($voucherCode, $expiryDate) {
		return ClassRegistry::init('BuzzBookings.BookingApi')->checkVoucherCode(
			$voucherCode,
			$expiryDate
		);
	}

/**
 * API call to check third-party vouchers (to enable unit test mocking).
 *
 * @param string $voucherCode
 * @param string $expiryDate Expiry date in SQL format
 * @return array
 */
	protected function _checkThirdPartyVoucherCodeApiCall($voucherCode, $expiryDate) {
		return ClassRegistry::init('BuzzBookings.BookingApi')->checkThirdPartyVoucherCode(
			$voucherCode,
			$expiryDate
		);
	}

/**
 * API call (to enable unit test mocking).
 *
 * @param int $activityId
 * @param int $voucherProviderId
 * @param string $voucherCode
 * @return bool
 */
	protected function _checkVoucherAvailable($activityId, $voucherProviderId, $voucherCode) {
		return ClassRegistry::init('BuzzBookings.BookingApi')->checkVoucherAvailable(
			$activityId,
			$voucherProviderId,
			$voucherCode
		);
	}

/**
 * Returns matching voucher data retrieved from the API for a given voucher.
 *
 * @return array|bool Voucher data or false if not found
 */
	protected function _getVoucherData($voucherProviderId, $apiVoucherId) {
		return CakeSession::read('BuzzBookings.vouchers.' . $voucherProviderId . 'Voucher' . $apiVoucherId);
	}

/**
 * Returns the API's voucher provider.
 *
 * @return int
 */
	public function getApiVoucherProvider() {
		return (int)Configure::read('BuzzBookings.voucher_provider_id');
	}

/**
 * Returns voucher data.
 *
 * @param int $activityId Activity ID
 * @return array
 */
	public function getVouchers($activityId) {
		// Get the vouchers.
		$data = ClassRegistry::init('BuzzBookings.BookingApi')->getVouchersBySite(
			Configure::read('api.site_id'),
			$activityId
		);
		// Save the voucher data to the session for checking against when
		// adding a voucher to the booking.
		CakeSession::write('BuzzBookings.vouchers', $data);
		// Return the voucher data.
		return $data;
	}

/**
 * Get the available dates for the booking.
 *
 * @param string $fromDate From date
 * @param string $toDate To date which the method will modify according to the booking
 * @return array
 */
	public function getAvailableDates($fromDate, $toDate) {
		// Get the current basket.
		$basket = $this->getBasket();

		// Modify the toDate if any vouchers expire before the toDate passed
		// to the method.
		$expiryDates = Hash::extract($basket['BookingItem'], '{n}.BookingItemVoucher.{n}.expiry_date');
		if (!empty($expiryDates)) {
			sort($expiryDates);
			$toDate = $expiryDates[0] < $toDate ? $expiryDates[0] : $toDate;
		}

		// Get the times and dates available for the booking.
		$data = ClassRegistry::init('BuzzBookings.BookingApi')->getDates(
			Configure::read('api.site_id'),
			$fromDate,
			$toDate,
			$this->_getBookingActivitiesList($basket['Booking']['id'])
		);

		return $data;
	}

/**
 * Returns an array of unavailables dates from the specified date range.
 *
 * @param array $dates
 * @param string $fromDate
 * @param string $toDate
 * @return array
 */
	public function calculateUnavailableDates(array $dates, $fromDate, $toDate) {
		$availableDates = [];
		$unavailableDates = [];

		foreach ($dates as $date) {
			$availableDates[$date['date']] = true;
		}

		$fromDate = strtotime($fromDate);
		$toDate = strtotime($toDate);
		for ($date = $fromDate; $date <= $toDate; $date += 86400) {
			$dateKey = date('d/m/Y', $date);
			if (isset($availableDates[$dateKey]) === false) {
				$unavailableDates[$dateKey] = true;
			}
		}

		return $unavailableDates;
	}

/**
 * Returns a list of API activities and quantities.
 *
 * @param int $bookingId Booking/Basket ID
 * @return array
 */
	protected function _getBookingActivitiesList($bookingId) {
		$result = $this->BookingItem->BookingItemPlace->find(
			'all',
			array(
				'joins' => array(
					array(
						'table' => 'booking_items',
						'alias' => 'BookingItem',
						'type' => 'INNER',
						'conditions' => 'BookingItem.id = BookingItemPlace.booking_item_id'
					)
				)
			)
		);

		$data = [];

		$items = Hash::combine(
			$result,
			'{n}.BookingItemPlace.booking_item_id',
			'{n}.BookingItemPlace.quantity',
			'{n}.BookingItemPlace.api_reference'
		);
		foreach ($items as $apiRef => $item) {
			$data[$apiRef] = array_sum($item);
		}

		return $data;
	}

/**
 * Returns non-bookable extras.
 *
 * @param array $basket Current basket
 * @param int $limit Maximum number of extras to return
 * @return array
 */
	public function getExtras(array $basket, $limit = 3) {
		$packages = Hash::extract($basket['BookingItem'], '{n}.activity_package_id');
		$result = $this->BookingExtra->Extra->getAvailableExtras(
			$basket['Booking']['activity_id'],
			Hash::extract($basket['BookingItem'], '{n}.activity_package_id'),
			$limit
		);
		return $result;
	}

/**
 * Add all upsells to the basket.
 *
 * @param array $bookingExtras
 * @param array $bookingItems
 * @return bool
 */
	public function addUpsells(array $bookingExtras, array $bookingItems) {
		$bookingId = $this->getBasketId();

		// Add upsells
		$result = $this->_addExtras($bookingId, $bookingExtras);
		if ($result === true) {
			// @todo add booking items
		}

		// Update the basket total.
		$this->_updateBasketTotal();

		return $result;
	}

/**
 * Add extras to the basket.
 *
 * @param int $bookingId
 * @param array $bookingExtras
 * @return bool
 */
	protected function _addExtras($bookingId, array $bookingExtras) {
		// Make sure we don't save items with zero quantity to the database.
		foreach ($bookingExtras as $key => $val) {
			if ((int)$val['quantity'] === 0) {
				unset($bookingExtras[$key]);
			}
		}

		// Get all the extras being purchased from the database so that we can
		// calculate the costs.
		$extras = $this->BookingExtra->Extra->find(
			'all',
			array(
				'conditions' => array(
					'Extra.id' => Hash::extract($bookingExtras, '{n}.extra_id')
				)
			)
		);
		$extras = Hash::combine($extras, '{n}.Extra.id', '{n}.Extra');
		foreach ($bookingExtras as &$bookingExtra) {
			$bookingExtra = $this->_generateBookingExtra($bookingId, $extras, $bookingExtra);
		}

		$data = array(
			'Booking' => array(
				'id' => $bookingId,
			),
			'BookingExtra' => $bookingExtras
		);

		return $this->saveAssociated($data, ['deep' => true]) !== false;
	}

/**
 * Generates a booking extra to attach to a booking
 *
 * @param int $bookingId
 * @param array $extras
 * @param array $bookingExtra
 * @return array
 */
	protected function _generateBookingExtra($bookingId, array $extras, array $bookingExtra) {
		$extra = $extras[$bookingExtra['extra_id']];

		return array(
			'id' => $bookingExtra['id'],
			'booking_id' => $bookingId,
			'api_reference' => $extra['api_reference'],
			'extra_unit_cost' => $extra['price'],
			'quantity' => $bookingExtra['quantity'],
			'total' => $bookingExtra['quantity'] * $extra['price'],
		);
	}

}
